<?php

namespace Drupal\rules_http_client\Plugin\RulesAction;

use Drupal\Component\Render\FormattableMarkup;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Messenger\MessengerTrait;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Utility\Error;
use Drupal\rules\Core\Annotation\RulesAction;
use Drupal\rules\Context\ContextDefinition;
use Drupal\rules\Core\RulesActionBase;
use GuzzleHttp\BodySummarizer;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\Exception\RequestException;
use GuzzleHttp\MessageFormatter;
use GuzzleHttp\Psr7\Request;
use Psr\Http\Message\RequestInterface;
use Psr\Http\Message\ResponseInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides "Rules Http client" rules action.
 *
 * @RulesAction(
 *   id = "rules_http_client",
 *   label = @Translation("Request HTTP data"),
 *   category = @Translation("Data"),
 *   context_definitions = {
 *     "url" = @ContextDefinition("string",
 *       label = @Translation("URL"),
 *       description = @Translation("Url address where to post, get and delete request send."),
 *       required = TRUE,
 *       multiple = TRUE,
 *     ),
 *     "headers" = @ContextDefinition("string",
 *       label = @Translation("Headers"),
 *       description = @Translation("Request headers to send as 'name: value' pairs, one per line (e.g., Accept: text/plain). See <a href='https://www.wikipedia.org/wiki/List_of_HTTP_header_fields'>wikipedia.org/wiki/List_of_HTTP_header_fields</a> for more information."),
 *       required = FALSE,
 *      ),
 *     "method" = @ContextDefinition("string",
 *       label = @Translation("Method"),
 *       description = @Translation("The HTTP request methods like'HEAD','POST','PUT','DELETE','TRACE','OPTIONS','CONNECT','PATCH' etc."),
 *       required = FALSE,
 *     ),
 *     "data" = @ContextDefinition("string",
 *       label = @Translation("Data"),
 *       description = @Translation("The request body, formatter as 'param=value&param=value&...' or one 'param=value' per line.."),
 *       required = FALSE,
 *       multiple = TRUE,
 *       assignment_restriction = "data",
 *     ),
 *     "max_redirects" = @ContextDefinition("integer",
 *       label = @Translation("Max Redirect"),
 *       description = @Translation("How many times a redirect may be followed."),
 *       default_value = 3,
 *       required = FALSE,
 *       assignment_restriction = "input",
 *     ),
 *     "timeout" = @ContextDefinition("float",
 *       label = @Translation("Timeout"),
 *       description = @Translation("The maximum number of seconds the request may take.."),
 *       default_value = 30,
 *       required = FALSE,
 *     ),
 *     "debug" = @ContextDefinition("boolean",
 *       label = @Translation("Debug"),
 *       description = @Translation("Show debug information for the request. Specify 'true' or 'false'."),
 *       default_value = "false",
 *       required = FALSE,
 *       assignment_restriction = "input",
 *     ),
 *   },
 *   provides = {
 *     "http_response" = @ContextDefinition("string",
 *       label = @Translation("HTTP data")
 *     )
 *   }
 * )
 *
 * @todo: Define that message Context should be textarea comparing with
 *   textfield Subject
 * @todo: Add access callback information from Drupal 7.
 */
class RulesHttpClient extends RulesActionBase implements ContainerFactoryPluginInterface {

  use MessengerTrait;

  /**
   * The current user.
   *
   * @var \Drupal\user\UserInterface
   */
  protected $currentUser;

  /**
   * The logger for the rules channel.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * The HTTP client to fetch the feed data with.
   *
   * @var \GuzzleHttp\ClientInterface
   */
  protected $httpClient;

  /**
   * The Rules HTTP client config.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected $rulesHttpConfig;

  /**
   * The body summarizer.
   *
   * @var \GuzzleHttp\BodySummarizerInterface
   */
  protected $bodySummarizer;

  /**
   * Constructs a httpClient object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin ID for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Session\AccountInterface $current_user
   *    The current user.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   *   The logger factory service.
   * @param \GuzzleHttp\ClientInterface $http_client
   *   The guzzle http client instance.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The config factory.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, AccountInterface $current_user, LoggerChannelFactoryInterface $logger_factory, ClientInterface $http_client, ConfigFactoryInterface $config_factory) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->currentUser = $current_user;
    $this->logger = $logger_factory->get('rules_http_client');
    $this->httpClient = $http_client;
    $this->rulesHttpConfig = $config_factory->get('rules_http_client.settings');
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('current_user'),
      $container->get('logger.factory'),
      $container->get('http_client'),
      $container->get('config.factory')
    );
  }

  /**
   * Send a system email.
   *
   * @param string[] $url
   *   Url addresses HTTP request.
   * @param string|null $headers
   *   (optional) Header information of HTTP Request.
   * @param string $method
   *   (optional) Method of HTTP request.
   * @param string|null $data
   *   (optional) Raw data of HTTP request.
   * @param int|null $max_redirects
   *   (optional) Max redirect for HTTP request.
   * @param int|null $timeout
   *   (optional) Time Out for HTTP request.
   * @param string|null $debug
   *  (optional) Whether to enable debug information.
   */
  protected function doExecute(array $url, $headers, $method, $data = NULL, $max_redirects = 3, $timeout = 30, $debug = 'false') {
    $debug = filter_var($debug, FILTER_VALIDATE_BOOLEAN);
    // Headers section.
    if ( !is_array($headers) ) {
      $headers = explode("\n", $headers);
    }
    if (is_array($headers)) {
      foreach ($headers as $header) {
        if (!empty($header) && strpos($header, ':') !== FALSE) {
          [$name, $value] = explode(':', $header, 2);
          if (!empty($name)) {
            $options['headers'][$name] = ltrim($value);
          }
        }
      }
    }
    $finalArray = [];
    if (is_array($data)) {
      // Data section.
      foreach ($data as $singleArray) {
        $finalSingleArray = explode('=', $singleArray);
        $finalArray[$finalSingleArray[0]] = $finalSingleArray[1];
      }

      // Json decode array.
      $finalArray = json_encode($finalArray);
    }

    // Payload data.
    $options['data'] = $finalArray;

    // Max redirects.
    $options['max_redirects'] = empty($max_redirects) ? 3 : $max_redirects;

    // Timeout.
    $options['timeout'] = empty($timeout) ? 30 : $timeout;

    $postUrl = $url[0];
    // Method.
    $method = strtoupper($method ?: 'GET');
    // Options.
    $options['method'] = $method;

    $options['body'] = $finalArray;

    $request = new Request($method, $postUrl, $options['headers'] ?? [], $options['body']);

    try {
      $response = $this->httpClient->send($request, $options);
      // Status of request.
      $status = $response->getStatusCode();
      // Check if we successfully get the output.
      if ($status >= 200 && $status < 300) {
        $stream = $response->getBody();
        $stream->rewind();
        $output = $stream->getContents();
        // Set the response output from service call.
        $this->setProvidedValue('http_response', $output);
      }
      if ($debug) {
        // Log the request.
        [$message, $variables] = $this->getDebugLogMessage($request, $response);
        $this->logger->debug($message, $variables);
        if ($this->rulesHttpConfig->get('show_responses') && $this->currentUser->hasPermission('access rules debug')) {
          // Show the request details to the user.
          $this->messenger()->addStatus(new FormattableMarkup("<h3>Rules HTTP Client request details</h3>\n$message", $variables));
        }
      }
    }
    catch (RequestException $error) {
      $request = $error->getRequest();
      $response = $error->getResponse();
      if ($debug) {
        [$message, $variables] = $this->getDebugLogMessage($request, $response);
        $this->logger->error($message, $variables);
        if ($this->rulesHttpConfig->get('show_responses') && $this->currentUser->hasPermission('access rules debug')) {
          // Show the request details to the user.
          $this->messenger()->addError(new FormattableMarkup("<h3>Rules HTTP Client request details</h3>\n$message", $variables));
        }
      }
      else {
        $error = RequestException::create($error->getRequest(), $error->getResponse(), $error, $error->getHandlerContext(), $this->getBodySummarizer());
        $variables = Error::decodeException($error);
        unset($variables['backtrace'], $variables['exception'], $variables['severity_level'], $variables['@backtrace_string']);
        $this->logger->error('%type Error details are as follows:<pre>@message</pre>', $variables);
      }
    }
    catch (\Exception $error) {
      $message = '%type while trying to connect to the remote API %post_url: @message in %function (line %line of %file)';
      $variables = Error::decodeException($error);
      $variables['%post_url'] = $postUrl;
      $this->logger->error($message, $variables);
    }
  }

  /**
   * Returns the debug log message.
   */
  protected function getDebugLogMessage(RequestInterface $request, ResponseInterface $response) {
    $template = [
      '<p><strong>Request details:</strong></p>',
      '<p>Method: <code>@request_method</code></p>',
      '<p>URL: <code>@request_url</code></p>',
      '<p>Headers: <pre><code>@request_headers</code></pre></p>',
      '<p>Body: <pre><code>@request_body</code></pre></p>',
      '<hr/>',
      '<p><strong>Response details:</strong></p>',
      '<p>Headers: <pre><code>@response_headers</code></pre></p>',
      '<p>Body: <pre><code>@response_body</code></pre></p>',
    ];
    $formatter = new MessageFormatter(implode("\n", $template));
    $message = $formatter->format($request, $response);

    $uri = $request->getUri();
    $user_info = $uri->getUserInfo();
    if (FALSE !== ($pos = \strpos($user_info, ':'))) {
      // Hide the basic auth password if any.
      $uri = $uri->withUserInfo(\substr($user_info, 0, $pos), '***');
    }
    $variables = [
      '@request_method' => $request->getMethod(),
      '@request_url' => (string) $uri,
      '@request_headers' => (new MessageFormatter('{req_headers}'))->format($request, $response),
      '@request_body' => (new MessageFormatter('{req_body}'))->format($request, $response),
      '@response_headers' => (new MessageFormatter('{res_headers}'))->format($request, $response),
      '@response_body' => (new MessageFormatter('{res_body}'))->format($request, $response),
      // '@response_body' => $this->getBodySummarizer()->summarize($response),
    ];
    return [$message, $variables];
  }

  /**
   * Returns the request body summarizer that isn't truncated at 120 characters.
   */
  protected function getBodySummarizer() {
    if ($this->bodySummarizer === NULL) {
      $this->bodySummarizer = new BodySummarizer($this->rulesHttpConfig->get('max_response_size') ?: 10000);
    }
    return $this->bodySummarizer;
  }

}
